/*
 This file is part of the digitalSTROM ds485-core library.

 Copyright (C) 2010 digitalSTROM.org, Schlieren, Switzerland

 digitalSTROM ds485-core library is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation version 2
 of the License.

 digitalSTROM ds485-core library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with digitalSTROM ds485-core library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

/**
 * @file
 * Provides the interface functions as defined in @see ds485-interface.h
 * Contains also some functions for ds485 sending/receiving
 */
#include "config.h"
#include "ds485-unix.h"

#include <assert.h>
#include <stdio.h>
#include <stdarg.h>
#include <signal.h>
#include <time.h>
#include <errno.h>
#include <pthread.h>

#include "digitalSTROM/ds485-const.h"
#include "digitalSTROM/ds485-core.h"
#include "digitalSTROM/ds485.h"

#include "ds485-qp.h"
#include "ds485-signals.h"
#include "ds485-proto.h"

#include "qp_port.h"
#include "receiver.h"
#include "logger.h"
#include "callqueue.h"
#include "../mmi.h"
#include "dsuid_helper.h"

Q_DEFINE_THIS_MODULE(ds485-unix)

/************************** STATIC DATA ****************************/

Ds485LibInst_t  devInst;
static bool isOpen   = FALSE;
extern QEQueue l_ds485RxMessageQueue;

void ds485_receiver(unsigned char *rxBuf, unsigned int rxLen);


/************************** dS485 SYSTEM ***************************/

void ds485_receiver(unsigned char *rxBuf, unsigned int rxLen)
{
    while (rxLen > 0) {
        rxLen--;

        switch (*rxBuf) {
        case DS485_START_BYTE: {
            static QEvent const startByteEvt = { RX_START_BYTE_SIG, 0 };
            QActive_postFIFO(AO_Receiver, &startByteEvt);
            break;
        }
        case DS485_ESCAPE_BYTE: {
            static QEvent const escapeByteEvt = { RX_ESCAPE_BYTE_SIG, 0 };
            QActive_postFIFO(AO_Receiver, &escapeByteEvt);
            break;
        }
        case DS485_RESERVED_BYTE: {
            static QEvent const reservedByteEvt = { RX_RESERVED_BYTE_SIG, 0 };
            QActive_postFIFO(AO_Receiver, &reservedByteEvt);
            break;
        }
        default: {
            OtherByteEvt* otherByteEvt = Q_NEW(OtherByteEvt, RX_OTHER_BYTE_SIG);
            otherByteEvt->byte = (uint8_t)*rxBuf;
            QActive_postFIFO(AO_Receiver, (QEvent*)otherByteEvt);
            break;
        }}
        rxBuf++;
    }
}


/**
 * Called by the dS485 protocol stack to immediately send a
 * frame on the wire
 */
int ds485_system_tx_raw(unsigned char *data, unsigned int length)
{
    Ds485LibInst_t* dev = &devInst;
    int result = 0;

    if (dev->tty.devfd >= 0) {
        if (write(dev->tty.devfd, data, length) != length) {
            result = -1;
        }
        tcdrain(dev->tty.devfd);
    }

    if (dev->serialLoopback) {
        ds485_receiver(data, length);

        if (dev->logFd >= 0) {
            if (write(dev->logFd, data, length) != length) {
                result = -1;
            }
        }
    }

    return result;
}

/*
 * called by protocol layer to give transmission control to
 * the system send queue
 * NOTE: this is a cooperative function, the system implementation
 *       must not block the token for too long!
 * NOTE: the token is automatically forwarded by the calling protocol
 *       layer when this function returns
 */
int ds485_system_tx()
{
    DS485LongMessageEvt const * txMsg;
    uint8_t* p;
    uint16_t length, n;
    static uint8_t data[1 + DS485_LONG_MESSAGE_SIZE * 2];

    txMsg = (DS485LongMessageEvt const *)DS485_GetNextTxMessage();

    if (txMsg == NULL) {
        /* this holds because it is exactly one message in the queue */
        Q_ERROR();
    }

    p = data;
    *p++ = 0xFF; /* add 0xFF to block noise */
    *p++ = DS485_START_BYTE;
    for (n = 1; n < txMsg->size; n++) {
        switch (txMsg->message[n]) {
        case DS485_ESCAPE_BYTE:
        case DS485_START_BYTE:
        case DS485_RESERVED_BYTE:
            *p++ = DS485_ESCAPE_BYTE;
            *p++ = txMsg->message[n] & 0x7f;
            break;
        default:
            *p++ = txMsg->message[n];
            break;
        }
    }

    /* explicitly recycle the event */
    QF_gc((QEvent const *) txMsg);

    length = p - data;
    return ds485_system_tx_raw(data, length);
}

static void* sysRxThread(void* arg)
{
    Ds485LibInst_t *dev = (Ds485LibInst_t *) arg;
    int rnum;
    int devOpenError = 0;
    uint8_t rxBuffer[512];

    /*
     * Main serial device handler
     */
    while (dev->mainActive) {

        if (strncmp("/dev", dev->devName, 4) == 0) {
           open_tty(dev->devName,&dev->tty,dev);
        } else {
            dev->tty.devfd = open(dev->devName, 0);
        }
        if (dev->tty.devfd < 0) {
            if (!devOpenError) {
                ds485_system_msg(dev, LOG_ERR,
                                 "ERROR: cannot open %s [%d]: error = %s (%d)\n",
                                 dev->devName, dev->tty.devfd,
                                 strerror(errno), errno);
            }
            sleep(1);
            devOpenError++;
            continue;
        }
        devOpenError = 0;

        ds485_system_msg(dev, LOG_DEBUG,
                         "Serial device open %s: %d\n",
                         dev->devName, dev->tty.devfd);

        while (dev->mainActive && dev->tty.devfd >= 0) {
            fd_set readfs;
            struct timeval Timeout;
            int rval;

            FD_ZERO(&readfs);
            FD_SET(dev->tty.devfd, &readfs);
            Timeout.tv_usec = 0; /* microseconds */
            Timeout.tv_sec = 1; /* seconds */

            rval = select(dev->tty.devfd + 1, &readfs, NULL, NULL, &Timeout);

            /* timeout */
            if (rval == 0) {
                continue;
            } else if (rval < 0) {
                /* error */
                break;
            }

            if (FD_ISSET(dev->tty.devfd, &readfs)) {
                /* data */
                rnum = read(dev->tty.devfd, rxBuffer, sizeof(rxBuffer));
                if (rnum == 0) {
                    if (errno == 0) {
                        ds485_system_msg(dev, LOG_DEBUG, "Serial port disconnected\n");
                        break;
                    } else if (errno == 2) {
                        sleep(2);
                        break;
                    }
                } else if (rnum < 0) {
                    ds485_system_msg(dev, LOG_ERR,
                                     "Serial port read error: %s (%d)\n",
                                     strerror(errno), errno);
                    break;
                } else if (rnum > 0) {
                    ds485_receiver(rxBuffer, rnum);
                    if (dev->logFd >= 0) {
                        if (write(dev->logFd, rxBuffer, rnum) != rnum) {
                            ds485_system_msg(dev, LOG_ERR,
                                             "Loopback write error\n");
                        }
                    }
                }
            }
        }

        ds485_system_msg(dev, LOG_DEBUG,"Serial port closed %d\n",
                         dev->tty.devfd);
        close(dev->tty.devfd);
        dev->tty.devfd = -1;
    }

    return NULL;
}

/* TODO civilize this hack */
extern void libqp_init(void);
extern void libqp_thread(void);

static void* sysQpThread(void* arg)
{
    ds485_system_msg(&devInst, LOG_DEBUG, "Startup QP/C libmain\n");
    libqp_thread();
    ds485_system_msg(&devInst, LOG_DEBUG, "Shutdown QP/C libmain\n");
    return NULL;
}

/********************* APPLICATION INTERFACE ***********************/

ds485Handle_t ds485_api_open(const char *device, const dsid_t *dsid)
{
    /*
     * Exclusive device
     */
    if (isOpen) {
        return NULL;
    }
    isOpen = TRUE;

    /*
     * Ressource allocation
     */
    Ds485LibInst_t* dev = &devInst;

    if (dev == NULL) {
        return NULL;
    }

    sem_init(&dev->semRx, 0, 0);
    sem_init(&dev->semTx, 0, 0);

    memset(dev, 0, sizeof(Ds485LibInst_t));

    dev->logFd = -1;
    dev->tty.devfd = -1;
    dev->devName = strdup(device);

    /*
     * Linux platform init
     */
    srandom(time(0));
    libqp_init();

    /*
     * start threads
     */
    dev->mainActive = 1;
    pthread_mutex_init(&dev->logMutex, NULL);

    pthread_attr_t attr;
    struct sched_param param;
    int rval;

    pthread_attr_init(&attr);
    param.sched_priority = 90;

    /* TODO: use logger service to emit errors */
    rval = pthread_attr_setinheritsched(&attr, PTHREAD_EXPLICIT_SCHED);
    if (rval) {
        fprintf(stderr, "ds485-core: failed to set QP/C thread attr explicit [%d]\n", rval);
    }
    rval = pthread_attr_setschedpolicy(&attr, SCHED_FIFO);
    if (rval) {
        fprintf(stderr, "ds485-core: failed to set QP/C thread policy [%d]\n", rval);
    }
    rval = pthread_attr_setschedparam(&attr, &param);
    if (rval) {
        fprintf(stderr, "ds485-core: failed to set QP/C thread priority to %d [%d]\n",
            param.sched_priority, rval);
    }

    if ((rval = pthread_create(&dev->qpThread, &attr, sysQpThread, dev)) != 0) {
        ds485_system_msg(&devInst, LOG_ERR,
            "ds485-core: failed to start QP/C thread [%d]\n", rval);
        return NULL;
    }
    pthread_attr_destroy(&attr);

    pthread_attr_init(&attr);
    param.sched_priority = 99;

    rval = pthread_attr_setinheritsched(&attr, PTHREAD_EXPLICIT_SCHED);
    if (rval) {
        fprintf(stderr, "ds485-core: failed to set receiver thread attr explicit [%d]\n", rval);
    }
    rval = pthread_attr_setschedpolicy(&attr, SCHED_FIFO);
    if (rval) {
        fprintf(stderr, "ds485-core: failed to set receiver thread policy [%d]\n", rval);
    }
    rval = pthread_attr_setschedparam(&attr, &param);
    if (rval) {
        fprintf(stderr, "ds485-core: failed to set receiver thread priority to %d [%d]\n",
            param.sched_priority, rval);
    }

    if ((rval = pthread_create(&dev->rxThread, &attr, sysRxThread, dev)) != 0) {
        ds485_system_msg(&devInst, LOG_ERR,
            "ds485-core: failed to start serial receiver thread [%d]\n", rval);
        return NULL;
    }
    pthread_attr_destroy(&attr);

    DatamodelConfigEvt *evt = Q_NEW(DatamodelConfigEvt, CONFIG_INIT_SIG);
    memcpy(&evt->dsmid, dsid, sizeof(dsid_t));
    QActive_postFIFO(AO_DS485, (QEvent *)evt);

    ccCreate(&dev->reqBusMembers, dev->reqBusMembersSto,
             Q_DIM(dev->reqBusMembersSto));
    ccCreate(&dev->reqBusState, dev->reqBusStateSto,
             Q_DIM(dev->reqBusStateSto));
    return dev;
}

int ds485_api_close( ds485Handle_t handle )
{
    Ds485LibInst_t* dev = handle;

    if (!isOpen) {
        return  -1;
    }

    /*
     * Stop threads & timer
     */
    dev->mainActive = 0;
    QF_stop();

    ccFree(&dev->reqBusMembers);
    ccFree(&dev->reqBusState);

    /*
     * Cleanup rx/tx threads
     */
    pthread_cancel(dev->rxThread);

    sem_destroy(&dev->semRx);
    sem_destroy(&dev->semTx);

    /*
     * Restore serial terminal settings
     */
    close_tty(&dev->tty,dev);

    /*
     * Cleanup memory ressources
     */
    free(dev->devName);
    isOpen = FALSE;

    return 0;
}

int ds485s_send(ds485Handle_t _handle, const dsuid_t *_dst, const dsuid_t *_src,
                DS485CommandType _type,
                const uint8_t *_buf,
                size_t _len)
{
    Ds485LibInst_t* __attribute__((unused)) dev = _handle;
    DS485LongMessageEvt *evt;
    uint8_t dst;

    assert(dev);
    if (_len > DS485_LONG_PAYLOAD_SIZE) {
        return DS485_INTERNAL_ERROR_ERANGE;
    }

    if (dsuid_is_broadcast(_dst)) {
        dst = DS485_BROADCAST_MASK;
    } else {
        dst = get_short_address(_dst);
        if (dst == DS485_INVALID_ADDRESS) {
            return DS485_INTERNAL_ERROR_INVALID_BUS_ADDR;
        }
    }

    /*
     * src is the short address of the daemon, inserted automatically
     * by the stack, see DS485_SetSource, deviceAddress
     */
    evt = DS485_CreateLongDataMessage(dst, 0, _type, _len);
    memcpy(&evt->message[DS485_DATA_OFFSET], _buf, _len);

    QActive_postFIFO(AO_DS485, (QEvent *)evt);
    return waitTxResponse();
}

int ds485s_recv(ds485Handle_t _handle, dsuid_t *_dst, dsuid_t *_src,
                DS485CommandType *_type,
                uint8_t *_buf,
                size_t len)
{
    Ds485LibInst_t* dev = _handle;
    DS485LongMessageEvt const * rxMsgEvt;
    int ret = 0;

    assert(dev);
    sem_wait(&dev->semRx);
    rxMsgEvt = (DS485LongMessageEvt const *)QEQueue_get(&l_ds485RxMessageQueue);
    if (NULL == rxMsgEvt) {
        return -1;
    }

    if (_dst) {
        if (DS485_IsBroadcast(rxMsgEvt)) {
            *_dst = DSUID_BROADCAST;
        } else {
            *_dst = *get_dsuid(DS485_GetDestination(rxMsgEvt));
        }
    }
    if (_src) {
        *_src = *get_dsuid(DS485_GetSource(rxMsgEvt));
    }
    if (_type) {
        *_type = DS485_GetCommandType(rxMsgEvt);
    }
    if (_buf) {
        if (len < DS485_GetPayloadLength(rxMsgEvt)) {
            return -DS485_INTERNAL_ERROR_ERANGE;
        }
        /*
         * dsm-api format: transactionId + payload
         * but for us it's all data
         */
        memcpy(_buf, &rxMsgEvt->message[DS485_DATA_OFFSET],
               DS485_GetPayloadLength(rxMsgEvt));
        ret = DS485_GetPayloadLength(rxMsgEvt);
    }

    QF_gc((QEvent const *)rxMsgEvt);
    return ret;
}

int ds485_api_ioctl(ds485Handle_t handle, unsigned int cmd, void* data,
                    unsigned int length )
{
    Ds485LibInst_t* dev = handle;
    int rval = 0;

    switch (cmd) {
    case IOC_SET_MSG_CB:
        dev->callback_msg_handler = (ds485_api_callback_msg_t) data;
        break;

    case IOC_SET_LOGGER_FD:
        dev->logFd    = *(int*) data;
        break;

    case IOC_SET_BUS_CHANGE_CB:
        dev->busChange = (ds485_api_callback_bus_change_t) data;
        break;

    case IOC_SET_BUS_STATE_CB:
        dev->state = (ds485_api_callback_state_t) data;
        break;

    case IOC_LOOPBACK_MODE:
        dev->serialLoopback = *(unsigned int *) data;
        break;
    }

    return rval;
}

int ds485_api_get_bus_members(ds485Handle_t handle, dsuid_t *members,
                              int maxNumberOfMembers)
{
    const DS485_DsuidsResponseEvt *res;
    Ds485LibInst_t* dev = handle;
    int i, nr;

    QActive_postFIFO(AO_DS485, ccPostRequest(&dev->reqBusMembers,
                                             DS485_REQ_BUS_MEMBERS_SIG));
    res = (const DS485_DsuidsResponseEvt *)
        ccWaitResponse(&dev->reqBusMembers);

    nr = 0;
    for (i = 0; i < DS485_MAX_SLAVE_ADDRESS + 1; i++ ){
        if ((nr < maxNumberOfMembers) && !dsuid_is_null(&res->dsuids[i])) {
            memcpy(members, &res->dsuids[i], sizeof(dsuid_t));
            members++;
            nr++;
        }
    }
    QF_gc((QEvent const *) res);
    return  nr;
}

ds485s_bus_state_t ds485_api_get_bus_state(ds485Handle_t handle)
{
    const MMI_BusStateEvt *res;
    Ds485LibInst_t* dev = handle;

    QActive_postFIFO( AO_MMI, ccPostRequest(&dev->reqBusState,
                                            MMI_REQ_BUS_STATE_SIG ));
    res = (const MMI_BusStateEvt*)ccWaitResponse(&dev->reqBusState);

    ds485s_bus_state_t state = res->state;
    QF_gc((QEvent const *)res);
    return state;
}

static const QEvent l_ds485RestartBusEvt = { DS485_TX_RESTART_BUS_SIG, 0 };

int ds485_api_restart_bus( ds485Handle_t handle )
{
    Ds485LibInst_t* dev = handle;

    if (dev == NULL) {
        return -1;
    }

    QActive_postFIFO(AO_DS485, &l_ds485RestartBusEvt);
    return waitTxResponse();
}
